<?php

namespace App;

use Nette\SmartObject;
use Nette\Utils\DateTime;

/**
 * @dbTable
 * @property int $id (autoIncrement)
 * @property int $utCreated
 * @property ?int $utValidFrom (?)
 * @property ?int $utValidTo (?)
 * @property decimal $creditCALeft
 * @property decimal $creditCLLeft
 * @property Customer $customer (fk)
 * @property Package $package (fk)
 * @property User $user (fk)
 * @property ?int $suspendValidLeft (?)
 * @property ?int $utValidToHint (?)
 * @related PrepaidPackageProduct $prepaidPackageProducts
 * @related ProductSell $productSells
 * @related PrepaidChange $prepaidChanges
 **/
class Prepaid extends BaseModel 
{
    /**
     * @var array
     */
    private $originalValues = [];

    public function getIdentify()
	{
		return $this->id.": ".$this->package->name;
	}

	public function getValidLeft()
	{
		return (!$this->instore && $this->utValidTo)
			? $this->utValidTo - time()
			: null;
	}

    public function getValidLeftHint()
    {
        return ($this->utValidToHint)
            ? $this->utValidToHint - time()
            : null;
    }

    public function getShowNotify()
	{
		$pa = $this->package; 
		return $pa->notifyBefore !== null && $this->validLeft <= $pa->notifyBefore;
	}

    public function getShowNotifyHint()
    {
        $pa = $this->package;
        return $pa->notifyBefore !== null && $this->getValidLeftHint() <= $pa->notifyBefore;
    }

    public function leftProductsByProductId($productId)
	{
		$ppid = PackageProduct::getBy(array(
			'packageId' => $this->packageId,
			'productId' => $productId
		))->id;
		return $this->leftProducts($ppid);
	}

	public function leftProducts($packageProductId) 
	{
		$row = $this->prepaidPackageProducts->where('packageProductId', $packageProductId)->limit(1)->fetch();
		return $row ? $row->amountLeft : 0;
		/* return $ls->count() 
			? $ls->fetch()->amountLeft
			: 0; */
	}

    const State_suspended = "suspended";
    const State_active = "active";
    const State_instore = "instore";
    const State_invalid = "invalid";
    const StateNames = [
        self::State_suspended => "pozastavená",
        self::State_active => "aktivní",
        self::State_instore => "neaktivní",
        self::State_invalid => "neplatená",
    ];

    public function getStatus() {
        if ($this->suspendValidLeft) return self::State_suspended;
        if ($this->utValidFrom == null) return self::State_instore;
        $t = time();
        return (($this->utValidTo == null) || ($this->utValidTo > $t))
                && ($this->utValidFrom < $t)
                && ($this->utValidFrom !== null)
            ? self::State_active
            : self::State_invalid;
    }

    public function getStatusName() {
        return self::StateNames[$this->getStatus()];
    }

	public function isInstore() 
	{
		// Glob::varbar(array($this->utValidFrom, ($this->utValidFrom == null)));
		return $this->utValidFrom == null;
	}
	
	public function isValid() 
	{
		$t = time();
		return 
			(($this->utValidTo == null) || ($this->utValidTo > $t))
			&& ($this->utValidFrom < $t)
			&& ($this->utValidFrom !== null);
	}
	
	public function addCredit(float $price, bool $creditCL)
	{
		if ($creditCL) $this->creditCLLeft += $price;
		else $this->creditCALeft += $price;
		$this->save();
	}

	public function useCredit(float $price, ?bool $creditNC = null) : CreditAmount
	{
		// První bezhotovostní, potom hotovostní, neďelit pokud je to možné, na sekyru je - hotově
		$res = new CreditAmount();
		$lastCA = $this->creditCALeft;
		$lastCL = $this->creditCLLeft;
		if ($this->creditCLLeft >= $price) 
		{ // Cele z bezhotovostního
			$res->creditCL = $price;
		}
		else if ($this->creditCALeft >= $price) 
		{ // Celé z hotovosti
			$res->creditCA = $price;
		}
		else 
		{
			// Dober bezhotovostní
			$res->creditCL = $this->creditCLLeft;
			// na sekyru z hotovosti
			$res->creditCA = $price - $res->creditCL;
		}
		$this->creditCALeft -= $res->creditCA;
		$this->creditCLLeft -= $res->creditCL;
		$res->creditExistsCA = max($lastCA, 0) - $this->creditCALeft;
		$res->creditExistsCL = max($lastCL, 0) - $this->creditCLLeft;
		$this->save();

		return $res; // $this->closeIfEmpty();
	}

	public function useProduct($packageProductId, $amount = 1) 
	{
		$pp = $this->prepaidPackageProducts->where(array('packageProductId' => $packageProductId))->fetch();
		$pp->amountLeft -= $amount;
		$pp->save();
		return false; // $this->closeIfEmpty();
	}

	public function isEmpty() 
	{
		// Kredit je > 0 neni prazdna
		if ($this->creditLeft != 0) return false;
		// Ma urcenou platnost
		if ($this->utValidTo !== null) return false;
		// Ma slevy?
		if ($this->package->products->where(array('priceDirect IS NOT NULL'))->count() > 0) return false;
		// Ma nevybrane produkty
		if ($this->productsLeft > 0) return false;
		
		return true;
	}
	
	public function getChanges() 
	{
		return $this->related('prepaidChange');
	}
	
	public function getProductsLeft() 
	{
		return $this->prepaidPackageProducts->where('amountLeft > 0')->sum('amountLeft');
	}
	
	public function closeIfEmpty() 
	{
		if ($this->empty) {
            $this->beforeChange();
			$this->utValidTo = time();
			$this->addChange('Platnost byla automaticky ukončena.');
			$this->save();
			return true;
		};
		return false;
	}

    public function getProps() {
        return $this->toArray([
            'utValidFrom',
            'utValidTo',
            'creditCALeft',
            'creditCLLeft',
            'suspendValidLeft',
            'utValidToHint'
        ]);
    }

    public function beforeChange() {
        $this->originalValues = $this->getProps();
    }

	public function addChange($change, $note = '') 
	{
		$ch = new PrepaidChange();
		$ch->change = $change;
		$ch->prepaidId = $this->id;
		$ch->note = $note;
		$ch->userId = Glob::currentAccountId();
		$ch->utChange = time();
        $propChanges = [];
        $props = $this->getProps();
        foreach (array_keys($props) as $k) {
            $propChanges[$k] = [
                isset($this->originalValues[$k]) ? $this->originalValues[$k] : null,
                $props[$k]
            ];
        }
        $ch->propChanges = $propChanges;
		$ch->save();
		return $ch;
	}

	public static function getCreditInfo($utFrom, $utTo) // : \object
	{
		// $res = [ 'total' => self::findAll()->sum('creditLeft')];
		// $row = $this->context->database->query('SELECT SUM');
		// return (object)$res;
		return self::repository()->newTable()
			->select("SUM(creditCALeft) AS totalCA, SUM(creditCLLeft) AS totalCL")
			->fetch();
	}

	public function getCreditLeft()
	{
		return $this->creditCALeft + $this->creditCLLeft;
	}
}

/**
 * @property float $credit
 */
class CreditAmount 
{
	use SmartObject;

	/** @var float */
	public $creditCA = 0;

	/** @var float */
	public $creditCL = 0;

	/** @var float */
	public $creditExistsCA = 0;

	/** @var float */
	public $creditExistsCL = 0;

	public function getCredit()
	{
		return $this->creditCA + $this->creditCL;
	}
}